/*********************************************************************/
/*                                                                   */
/* Program name: AMQSFIN                                             */
/*                                                                   */
/* Source name:  amqsfina.c                                          */
/*                                                                   */
/* Description:  Sample C program that subscribes to an MQSeries     */
/*               broker for all FinalScore publications published    */
/*               by the modified results server sample amqsres2.     */
/*                                                                   */
/*  Usage:       amqsfin <QMgrName>                                  */
/*                                                                   */
/*  Language:    C                                                   */
/*                                                                   */
/*********************************************************************/
/* Statement:    Licensed Materials - Property of IBM                */
/*                                                                   */
/*               SupportPac MA0E                                     */
/*               (C) Copyright IBM Corp. 1999                        */
/*                                                                   */
/*********************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* MQSeries header files */
#include <cmqc.h>                           /* MQI                   */
#include <cmqpsc.h>                         /* MQI Publish/Subscribe */

/*********************************************************************/
/* General definitions:                                              */
/*********************************************************************/
#define MAX_STRING_SIZE   100        /* maximum string length        */
#define MAX_TOPIC_LENGTH  256        /* maximum topic length         */
#define MAX_MSG_LENGTH    1024       /* maximum message size         */
#define MAX_WAIT_TIME     180000     /* period of inactivity         */
#define MAX_RESPONSE_TIME 6000       /* maximum response wait        */

/*********************************************************************/
/* Queue for publications and responses to be sent.                  */
/*********************************************************************/
#define SUBSCRIBER_QUEUE  "FINAL.SCORE.SAMPLE.QUEUE"

/*********************************************************************/
/* An MQRFH structure and MQMD that contains the default values for  */
/* all their fields.                                                 */
/*********************************************************************/
MQRFH DefaultMQRFH = { MQRFH_DEFAULT };
MQMD DefaultMQMD   = { MQMD_DEFAULT };

/*********************************************************************/
/* Prototypes                                                        */
/*********************************************************************/
void SendBrokerCommand( MQHCONN  hConn
                      , MQHOBJ   hBrkObj
                      , MQHOBJ   hSubObj
                      , char     Command[]
                      , PMQLONG  pCompCode
                      , PMQLONG  pReason );

void CheckResponse( MQHCONN   hConn
                  , MQHOBJ    hObj
                  , PMQBYTE   pPublishMsgId
                  , PMQLONG   pCompCode
                  , PMQLONG   pReason );

/*********************************************************************/
/* Functions:                                                        */
/*********************************************************************/

/*********************************************************************/
/*                                                                   */
/* Function Name : main                                              */
/*                                                                   */
/* Description   : Entry function of the sample, connects to the     */
/*                 queue manager, registers as a subscriber of the   */
/*                 topic, waits for publications to arrive,          */
/*                 displaying them when they do. After a period of   */
/*                 inactivity the subscription is deregistered and   */
/*                 the sample ends.                                  */
/*                                                                   */
/* Flow          :                                                   */
/*                                                                   */
/*   Verify arguments                                                */
/*   MQCONN to broker queue manager                                  */
/*    MQOPEN broker control queue                                    */
/*    MQOPEN subscriber queue                                        */
/*     Register as a subscriber to the topic...                      */
/*      Wait for publications                                        */
/*       Display publications                                        */
/*      End wait                                                     */
/*      Deregister the subscriber from the topic...                  */
/*    MQCLOSE broker control queue                                   */
/*    MQCLOSE subscriber queue                                       */
/*   MQDISC from broker queue manager                                */
/*                                                                   */
/* Input Parms   : int  argc                                         */
/*                  Number of arguments                              */
/*                 char *argv[]                                      */
/*                  Program Arguments                                */
/*                                                                   */
/*********************************************************************/
int main(int argc, char **argv)
{
  /* MQSeries variables */
  MQHCONN      hConn = MQHC_UNUSABLE_HCONN;
  MQHOBJ       hBrkObj = MQHO_UNUSABLE_HOBJ;
  MQHOBJ       hSubObj = MQHO_UNUSABLE_HOBJ;
  MQLONG       Options;
  MQLONG       CompCode;
  MQLONG       Reason;
  MQLONG       ConnReason;
  MQOD         od  = { MQOD_DEFAULT };
  MQGMO        gmo = { MQGMO_DEFAULT };
  MQMD         md  = { MQMD_DEFAULT };

  /* Supplied Arguments */
  char         QMName[MQ_Q_MGR_NAME_LENGTH+1] = ""; /* Queue Manager */

  /* Broker Message variables */
  char        *pUserData;
  PMQBYTE      pMessageBlock = NULL;
  MQLONG       messageLength;
  PMQRFH       pMQRFHeader;

  /*******************************************************************/
  /* If no queue manager name was given as an argument, connect to   */
  /* the default queue manager (if one exists). Otherwise connect    */
  /* to the one specified.                                           */
  /*******************************************************************/
  if (argc > 1)
    strcpy(QMName, argv[1]);

  /*******************************************************************/
  /* Connect to the queue manager.                                   */
  /*******************************************************************/
  MQCONN( QMName
        , &hConn
        , &CompCode
        , &ConnReason );
  if( CompCode == MQCC_FAILED )
  {
    printf("MQCONN failed with CompCode %d and Reason %d\n",
            CompCode, ConnReason);
  }
  /*******************************************************************/
  /* If the queue manager was already connected we can ignore the    */
  /* warning for now and continue.                                   */
  /*******************************************************************/
  else if( ConnReason == MQRC_ALREADY_CONNECTED )
  {
    CompCode = MQCC_OK;
  }

  /*******************************************************************/
  /* Open the Broker's control queue for registration                */
  /*******************************************************************/
  if( CompCode == MQCC_OK )
  {
    /*****************************************************************/
    /* Open The broker's control queue for putting command messages  */
    /* to.                                                           */
    /*****************************************************************/
    strncpy(od.ObjectName, "SYSTEM.BROKER.CONTROL.QUEUE" , (size_t)MQ_Q_NAME_LENGTH);
    Options = MQOO_OUTPUT + MQOO_FAIL_IF_QUIESCING;
    MQOPEN( hConn
          , &od
          , Options
          , &hBrkObj
          , &CompCode
          , &Reason );
    if( CompCode != MQCC_OK )
    {
      printf("MQOPEN failed to open \"%s\"\nwith CompCode %d and Reason %d\n",
             od.ObjectName, CompCode, Reason);
    }
  }

  /*******************************************************************/
  /* Open the subscriber queue for publications from the broker.     */
  /*******************************************************************/
  if( CompCode == MQCC_OK )
  {
    strncpy(od.ObjectName, SUBSCRIBER_QUEUE, (size_t)MQ_Q_NAME_LENGTH);
    Options = MQOO_INPUT_EXCLUSIVE + MQOO_FAIL_IF_QUIESCING;
    MQOPEN( hConn
          , &od
          , Options
          , &hSubObj
          , &CompCode
          , &Reason );
    if( CompCode != MQCC_OK )
    {
      printf("MQOPEN failed to open \"%s\"\nwith CompCode %d and Reason %d\n",
             od.ObjectName, CompCode, Reason);
    }
  }


  /*******************************************************************/
  /* Register us as a subscriber to the topic.                       */
  /*******************************************************************/
  if( CompCode == MQCC_OK )
  {
    /*****************************************************************/
    /* The SendBrokerCommand function can be found later in this     */
    /* sample file, it takes all paramaters it requires to form a    */
    /* command message to sent to the broker's control queue.        */
    /* The first command we need to send to the broker is a          */
    /* subscriber registration to tell the broker that we are        */
    /* interested in the supplied topic and wish all related         */
    /* publications to be sent to us.                                */
    /*****************************************************************/
    SendBrokerCommand( hConn
                     , hBrkObj
                     , hSubObj
                     , MQPS_REGISTER_SUBSCRIBER
                     , &CompCode
                     , &Reason );
  }      

  if( CompCode == MQCC_OK )
  {
    printf("Subscriber registered.\n");
    /*****************************************************************/
    /* Allocate a storage block for the messages to be received in.  */
    /*****************************************************************/
    messageLength = MAX_MSG_LENGTH;
    pMessageBlock = (PMQBYTE)malloc(messageLength);
    if( pMessageBlock == NULL )
    {
      printf("Unable to allocate storage\n");
    }
    else
    {
      /***************************************************************/
      /* Get with wait on the subscriber's queue to receive any      */
      /* publications sent to the subscriber and display the user    */
      /* data contained. End the wait after 60 seconds of inactivity.*/
      /***************************************************************/
      gmo.Options = MQGMO_WAIT + MQGMO_CONVERT;
      gmo.WaitInterval = MAX_WAIT_TIME;

      printf("Waiting for final scores...\n");

      /***************************************************************/
      /* While messages are still being received display them.       */
      /***************************************************************/
      while (CompCode != MQCC_FAILED)
      {
        /*************************************************************/
        /* Set the MQMD to its default settings before we get a      */
        /* message.                                                  */
        /*************************************************************/
        memcpy( &md, &DefaultMQMD, sizeof(MQMD));

        MQGET( hConn
             , hSubObj
             , &md
             , &gmo
             , MAX_MSG_LENGTH
             , pMessageBlock
             , &messageLength
             , &CompCode
             , &Reason );

        if (CompCode == MQCC_OK)
        {
          /***********************************************************/
          /* We are only expecting messages in the MQRFH format,     */
          /* Check that the message returned from MQGET is an MQRFH  */
          /* message.                                                */
          /***********************************************************/
          if( memcmp(md.Format, MQFMT_RF_HEADER , MQ_FORMAT_LENGTH) == 0 )
          {
            /*********************************************************/
            /* The only part of the publication that we are          */
            /* interested in in this sample is the string data that  */
            /* has been added to the message after the MQRFH and     */
            /* NameValueString structures by the amqsres2 sample.    */
            /*********************************************************/
            pMQRFHeader = (PMQRFH)pMessageBlock;
            pUserData = (char *)(pMessageBlock + pMQRFHeader->StrucLength);
            if( pMQRFHeader->StrucLength < messageLength )
            {
              printf( "%s\n", pUserData);
            }
          }
          else
          {
            printf("Unexpected message format: %.8s\n", md.Format );
            CompCode = MQCC_FAILED;
          }
        }
        /*************************************************************/
        /* The MQGET failed without timeing out. If the MQGET did    */
        /* time out (Reason = MQRC_NO_MSG_AVAILABLE) it meens we     */
        /* have not received a publication for 60 seconds so we can  */
        /* assume the publisher sample(s) have finished and we can   */
        /* deregister the subscription.                              */
        /*************************************************************/
        else if (Reason != MQRC_NO_MSG_AVAILABLE)
          printf("MQGET failed with reason=%ld\n", Reason);
      } /* end of while */
      /***************************************************************/
      /* Free the message block.                                     */
      /***************************************************************/
      free( pMessageBlock );
    }

    /*****************************************************************/
    /* Deregister the subscription to the final score publications.  */
    /*****************************************************************/
    SendBrokerCommand( hConn
                     , hBrkObj
                     , hSubObj
                     , MQPS_DEREGISTER_SUBSCRIBER
                     , &CompCode
                     , &Reason );

    if( CompCode == MQCC_OK )
      printf("Deregistered subscriber.\n");
  }

  /*******************************************************************/
  /* MQCLOSE the control queue used by this sample.                  */
  /*******************************************************************/
  if( hBrkObj != MQHO_UNUSABLE_HOBJ )
  {
    MQCLOSE( hConn
           , &hBrkObj
           , MQCO_NONE
           , &CompCode
           , &Reason );
    if( CompCode != MQCC_OK )
      printf("MQCLOSE failed with CompCode %d and Reason %d\n",
             CompCode, Reason);
  }

  /*******************************************************************/
  /* MQCLOSE the subscriber queue used by this sample.               */
  /*******************************************************************/
  if( hSubObj != MQHO_UNUSABLE_HOBJ )
  {
    MQCLOSE( hConn
           , &hSubObj
           , MQCO_NONE
           , &CompCode
           , &Reason );
    if( CompCode != MQCC_OK )
      printf("MQCLOSE failed with CompCode %d and Reason %d\n",
             CompCode, Reason);
  }

  /*******************************************************************/
  /* Disconnect from the queue manager only if the connection        */
  /* worked and we were not already connected.                       */
  /*******************************************************************/
  if( (hConn != MQHC_UNUSABLE_HCONN)
    &&(ConnReason != MQRC_ALREADY_CONNECTED) )
  {
    MQDISC( &hConn
          , &CompCode
          , &Reason );
    if( CompCode != MQCC_OK )
      printf("MQDISC failed with CompCode %d and Reason %d\n",
              CompCode, Reason);
  }

  return(0);
}
/*********************************************************************/
/* end of main                                                       */
/*********************************************************************/

/*********************************************************************/
/*                                                                   */
/* Function Name : SendBrokerCommand                                 */
/*                                                                   */
/* Description   : Form an MQRFH command message and put it to the   */
/*                 broker's control queue.                           */
/*                                                                   */
/* Flow          :                                                   */
/*                                                                   */
/*    Allocate the message block                                     */
/*     Define the MQRFH at the start of the message                  */
/*     Form the NameValueString following the MQRFH                  */
/*     MQPUT the message block to the control queue                  */
/*      Check the response from the broker...                        */
/*                                                                   */
/* Input Parms   : MQHCONN  hConn                                    */
/*                  Queue Manager connection handle                  */
/*                 MQHOBJ   hBrkObj                                  */
/*                  Object handle of control queue                   */
/*                 MQHOBJ   hSubObj                                  */
/*                  Object handle of subscriber queue (responses)    */
/*                 char     Command[]                                */
/*                  MQPS command string                              */
/*                                                                   */
/* Ouput Parms   : PMQLONG  pCompCode                                */
/*                  Pointer to completion code                       */
/*                 PMQLONG  pReason                                  */
/*                  Pointer to reason                                */
/*                                                                   */
/*********************************************************************/
void SendBrokerCommand( MQHCONN  hConn
                      , MQHOBJ   hBrkObj
                      , MQHOBJ   hSubObj
                      , char     Command[]
                      , PMQLONG  pCompCode
                      , PMQLONG  pReason )
{
 PMQBYTE pMessageBlock = NULL;
 MQLONG   messageLength;
 PMQRFH  pMQRFHeader;
 PMQCHAR  pNameValueString;
 MQLONG NameValueStringLength;
 MQMD  md = { MQMD_DEFAULT };
 MQPMO pmo = { MQPMO_DEFAULT };

  /*******************************************************************/
  /* Allocate a storage block for the command to be built in.        */
  /*******************************************************************/
  messageLength = MAX_MSG_LENGTH;
  pMessageBlock = (PMQBYTE)malloc(messageLength);
  if( pMessageBlock == NULL )
  {
    printf("Unable to allocate storage\n");
  }
  else
  {
    /*****************************************************************/
    /* Clear the buffer before we start (initialise to nulls).       */
    /*****************************************************************/
    memset(pMessageBlock, 0, MAX_MSG_LENGTH);

    /*****************************************************************/
    /* Define the MQRFH values.                                      */
    /*****************************************************************/
    pMQRFHeader = (PMQRFH)(pMessageBlock);
    memcpy( pMQRFHeader, &DefaultMQRFH, (size_t)MQRFH_STRUC_LENGTH_FIXED);

    /*****************************************************************/
    /* Start the NameValueString directly after the MQRFH            */
    /* structure and add the token pairs for the command.            */
    /*****************************************************************/
    pNameValueString = ((MQCHAR *)pMQRFHeader) + MQRFH_STRUC_LENGTH_FIXED;

    /*****************************************************************/
    /* Step 1.                                                       */
    /*****************************************************************/
    /* Now we must define the contents of the NameValueString, this  */
    /* is similar to the earlier exercise, amqssub. This time the    */
    /* stream name is not optional because we are NOT using the      */
    /* default stream, we are using the results service sample       */
    /* stream, the topic also differs, this sample will only ever    */
    /* subscribe to one topic.                                       */
    /* TIP: We pass in the MQPS command string to this function in   */
    /*      the string argument 'Command', insert this in the        */
    /*      NameValueString as the MQPS command value.               */
    /* TIP: The topic we subscribe to must match against all         */
    /*      publications from amqsres2 that are for a final score,   */
    /*      (which are in the form :                                 */
    /*                 "Sport/Soccer/State/FinalScore/Team1 Team2")  */
    /*      use a wildcard to subscribe to all matching publications.*/
    /* TIP: As before, you can use the string functions strcat and   */
    /*      strcpy to build up the NameValueString.                  */
    /*****************************************************************/
    < ADD CODE HERE >
    /*****************************************************************/
    /* End of Step 1.                                                */
    /*****************************************************************/
    /* This is the end of the exercises in this sample, return to    */
    /* the exercise sheet.                                           */
    /*****************************************************************/

    /*****************************************************************/
    /* Include the null terminator character in the                  */
    /* NameValueString. This is not required by the broker but it    */
    /* does allow us to use string functions on it in the future.    */
    /*****************************************************************/
    NameValueStringLength = strlen(pNameValueString) + 1;

    /*****************************************************************/
    /* Calculate the end of the MQRFH and NameValueString            */
    /* structure, it is recommended that this is alligned to a       */
    /* sixteen byte boundary.                                        */
    /*****************************************************************/
    messageLength = MQRFH_STRUC_LENGTH_FIXED 
                                 + ((NameValueStringLength+15)/16)*16;
    pMQRFHeader->StrucLength = messageLength;

    /*****************************************************************/
    /* Set the md for an MQRFH message. Send the message as a        */
    /* request message so that the broker always sends us a          */
    /* response to the reply queue.                                  */
    /*****************************************************************/
    memcpy(md.Format, MQFMT_RF_HEADER, MQ_FORMAT_LENGTH);
    md.MsgType = MQMT_REQUEST;
    strcpy( md.ReplyToQ, SUBSCRIBER_QUEUE);
    pmo.Options |= MQPMO_NEW_MSG_ID;
  
    MQPUT( hConn
         , hBrkObj
         , &md
         , &pmo
         , messageLength
         , pMessageBlock
         , pCompCode
         , pReason );

    if (*pCompCode != MQCC_OK)
    {
      printf("MQPUT failed with CompCode %d and Reason %d\n",
            *pCompCode, *pReason);
    }
    else
    {
      /***************************************************************/
      /* If the MQPUT succeeded check that an okay response is       */
      /* returned from the broker.                                   */
      /***************************************************************/
      CheckResponse( hConn
                   , hSubObj
                   , md.MsgId
                   , pCompCode
                   , pReason );
    }
    /*****************************************************************/
    /* Free the message block.                                       */
    /*****************************************************************/
    free( pMessageBlock );
  }
}
/*********************************************************************/
/* end of SendBrokerCommand                                          */
/*********************************************************************/

/*********************************************************************/
/*                                                                   */
/* Function Name : CheckResponse                                     */
/*                                                                   */
/* Description   : Get the response that should have been sent from  */
/*                 the broker on receiving the publication, if one   */
/*                 is received check for an okay response, if one    */
/*                 is not returned it may be because the broker      */
/*                 is not running.                                   */
/*                                                                   */
/* Flow          :                                                   */
/*                                                                   */
/*    Define the MQGMO and MQMD for the MQGET                        */
/*    MQGET with wait from the reply queue for the response          */
/*     Check if the response is okay                                 */
/*      If it is not display the error message                       */
/*    Return the CompCode to the calling function                    */
/*                                                                   */
/* Input Parms   : MQHCONN  hConn                                    */
/*                  Queue Manager connection handle                  */
/*                 MQHOBJ   hObj                                     */
/*                  Object handle of queue                           */
/*                 PMQBYTE  pPublishMsgId                            */
/*                  Pointer to MsgId of published message            */
/*                                                                   */
/* Ouput Parms   : PMQLONG  pCompCode                                */
/*                  Pointer to completion code                       */
/*                 PMQLONG  pReason                                  */
/*                  Pointer to reason                                */
/*                                                                   */
/*********************************************************************/
void CheckResponse( MQHCONN   hConn
                  , MQHOBJ    hObj
                  , PMQBYTE   pPublishMsgId
                  , PMQLONG   pCompCode
                  , PMQLONG   pReason )
{
  MQGMO        gmo   = { MQGMO_DEFAULT };
  MQMD         md    = { MQMD_DEFAULT };
  PMQBYTE      pMessageBlock = NULL;
  MQLONG       messageLength;
  PMQRFH       pMQRFHeader;
  PMQCHAR      pNameValueString;
  PMQCHAR      pInputNameValueString;

  /*******************************************************************/
  /* Allocate a storage block for the publications to be received    */
  /* into.                                                           */
  /*******************************************************************/
  messageLength = MAX_MSG_LENGTH;
  pMessageBlock = (PMQBYTE)malloc(messageLength);
  if( pMessageBlock == NULL )
  {
    printf("Unable to allocate storage\n");
    *pCompCode = MQCC_FAILED;
  }
  else
  {
    /*****************************************************************/
    /* Wait for a response message to arrive on our reply queue, the */
    /* response's correlId will be the same as the messageId that    */
    /* the original message was sent with (returned in the md from   */
    /* the MQPUT) so match against this.                             */
    /*****************************************************************/
    memcpy( md.CorrelId, pPublishMsgId, sizeof(MQBYTE24));
    gmo.Version = MQGMO_VERSION_2;
    gmo.MatchOptions = MQMO_MATCH_CORREL_ID;
    gmo.Options = MQGMO_WAIT + MQGMO_CONVERT;
    gmo.WaitInterval = MAX_RESPONSE_TIME;
  
    MQGET( hConn
         , hObj
         , &md
         , &gmo
         , messageLength
         , pMessageBlock
         , &messageLength
         , pCompCode
         , pReason );

    if( *pCompCode != MQCC_OK )
    {
      printf("MQGET failed with CompCode %d and Reason %d\n",
                                                 *pCompCode, *pReason);
      if( *pReason == MQRC_NO_MSG_AVAILABLE )
        printf("No response was sent by the broker, check the broker is running.\n");
    }
    else
    {
      /***************************************************************/
      /* Check that the message is in the MQRFH format.              */
      /***************************************************************/
      if( memcmp(md.Format, MQFMT_RF_HEADER, MQ_FORMAT_LENGTH) == 0 )
      {
        /*************************************************************/
        /* Locate the start of the NameValueString.                  */
        /*************************************************************/
        pMQRFHeader = (PMQRFH)pMessageBlock;
        pNameValueString = (PMQCHAR)(pMessageBlock 
                                            + MQRFH_STRUC_LENGTH_FIXED);

        /*************************************************************/
        /* The start of a response NameValueString is always in the  */
        /* same format:                                              */
        /*   MQPSCompCode x MQPSReason y MQPSReasonText string ...   */
        /* We can scan the start of the string to check the CompCode */
        /* and reason of the reply. The NameValueString of a response*/
        /* is always null terminated so we can use string functions. */
        /*************************************************************/
        sscanf(pNameValueString, "MQPSCompCode %d MQPSReason %d ",
                                                   pCompCode, pReason);
        if( *pCompCode != MQCC_OK )
        {
          /***********************************************************/
          /* A non-okay response was received, display the error     */
          /* message supplied along with the user data that was      */
          /* returned (if any), this will be the original command's  */
          /* NameValueString.                                        */
          /***********************************************************/
          printf("ERROR response returned :\n");
          printf(" %s\n",pNameValueString);
          /***********************************************************/
          /* Display the user data if supplied.                      */
          /***********************************************************/
          if( messageLength != pMQRFHeader->StrucLength )
          {
            printf("Original Command String:\n");
            pInputNameValueString = 
                   (PMQCHAR)(pMessageBlock + pMQRFHeader->StrucLength);
            /*********************************************************/
            /* We know that our original NameValueString was null    */
            /* terminated so we can simply printf the string.        */
            /*********************************************************/
            printf(" %s\n",pInputNameValueString);
          } 
        }
      }
      /***************************************************************/
      /* If the message is not in the MQRFH format we have the wrong */
      /* message (this should never happen).                         */
      /***************************************************************/
      else
      {
        printf("Unexpected message format: %.8s\n", md.Format );
        *pCompCode = MQCC_FAILED;
      }
    }
    /*****************************************************************/
    /* Free the message block.                                       */
    /*****************************************************************/
    free(pMessageBlock);
  }
}
/*********************************************************************/
/* end of CheckResponse                                              */
/*********************************************************************/

/*********************************************************************/
/*                                                                   */
/* end of amqssuba.c                                                 */
/*                                                                   */
/*********************************************************************/

                      
                      

